#include <fstream>
#include <math.h>

#include <microcpp/microcpp.h>
#include <rapidjson/document.h>
#include <rapidjson/istreamwrapper.h>
#include <rapidjson/error/en.h>

#include <engine.h>
#include <main.h>
#include <fighterinfo.h>

void printMsg(std::string text)
{
	screen::print(SCREEN_UP, 128-(text.size()*6/2), 96-4, text.c_str());
	screen::render();
}

void pause2(const char* text)
{
	for (int i=0; i<60; i++)
	{
		screen::print(0, 8, 8, text);
		screen::render();
	}
}

void skew(float* topleft, float* topright, float* botleft, float* botright, float amnt)
{
	(*topleft) += amnt;
	(*topright) += amnt;
	(*botleft) -= amnt;
	(*botright) -= amnt;
}


Engine::Engine()
{
	started = false;
	for (int i=0; i<2; i++)
	{
		players[i] = new Player(i, this);
	}
}

void Engine::setStage(std::string name)
{
	if (started) return;
	stage.name = name;
	stage.bgs.clear();

	loadActions();

	char* buf = new char[8192];

	sprintf(buf, "Setting stage to %s", stage.name.c_str());
	printMsg(buf);

	sprintf(buf, "/data/smbz-ds/stages/%s/stage.json", stage.name.c_str());
	rapidjson::Document doc;
	std::ifstream ifs(buf);
	rapidjson::IStreamWrapper isw(ifs);
	rapidjson::ParseResult ok = doc.ParseStream(isw);
	ifs.close();

	if (!ok)
	{
		sprintf(buf, "Stage %s JSON parse error: %s (%u)", name.c_str(), rapidjson::GetParseError_En(ok.Code()), ok.Offset());
		gameCrash(buf);
	}

	stage.bgcolor = (doc.HasMember("bgcolor")) ? Color::new256(doc["bgcolor"][0].GetInt(), doc["bgcolor"][1].GetInt(), doc["bgcolor"][2].GetInt()) : 0;

	stage.ground.on = (!doc["ground"].HasMember("on") || doc["ground"]["on"].GetBool());
	if (stage.ground.on)
	{
		stage.ground.offsetY = doc["ground"]["offsetY"].GetInt();
		stage.ground.topFrame.x = doc["ground"]["topFrame"]["x"].GetInt();
		stage.ground.topFrame.y = doc["ground"]["topFrame"]["y"].GetInt();
		stage.ground.topFrame.w = doc["ground"]["topFrame"]["w"].GetInt();
		stage.ground.topFrame.h = doc["ground"]["topFrame"]["h"].GetInt();
		stage.ground.loopFrame.x = doc["ground"]["loopFrame"]["x"].GetInt();
		stage.ground.loopFrame.y = doc["ground"]["loopFrame"]["y"].GetInt();
		stage.ground.loopFrame.w = doc["ground"]["loopFrame"]["w"].GetInt();
		stage.ground.loopFrame.h = doc["ground"]["loopFrame"]["h"].GetInt();
	}

	for (u32 i=0; i<doc["bgs"].Capacity(); i++)
	{
		BgInfo bg;

		// setup struct values
		bg.offsetX = doc["bgs"][i]["offset"]["x"].GetInt();
		bg.offsetY = doc["bgs"][i]["offset"]["y"].GetInt();
		bg.paraX = doc["bgs"][i]["parallax"]["x"].GetFloat();
		bg.paraY = doc["bgs"][i]["parallax"]["y"].GetFloat();
		bg.tileX = doc["bgs"][i]["tile"]["x"].GetBool();
		bg.tileY = doc["bgs"][i]["tile"]["y"].GetBool();
		bg.scale = (doc["bgs"][i].HasMember("scale") && doc["bgs"][i]["scale"].GetBool());

		if (doc["bgs"][i].HasMember("sep"))
		{
			bg.sepX = doc["bgs"][i]["sep"]["x"].GetInt();
			bg.sepY = doc["bgs"][i]["sep"]["y"].GetInt();
		}
		else
			bg.sepX = bg.sepY = 0;
		if (doc["bgs"][i].HasMember("scroll"))
		{
			bg.scrollX = doc["bgs"][i]["scroll"]["x"].GetFloat();
			bg.scrollY = doc["bgs"][i]["scroll"]["y"].GetFloat();
		}
		else
			bg.scrollX = bg.scrollY = 0;
		bg.scrollPosX = bg.scrollPosY = 0;

		if (doc["bgs"][i].HasMember("filename"))
		{
			bg.isRect = false;
			bg.filename = doc["bgs"][i]["filename"].GetString();
			if (doc["bgs"][i].HasMember("frames")) // animated background
			{
				for (u32 j=0; j<doc["bgs"][i]["frames"].Capacity(); j++)
				{
					BgFrame frame;
					frame.x = doc["bgs"][i]["frames"][j]["x"].GetInt();
					frame.y = doc["bgs"][i]["frames"][j]["y"].GetInt();
					frame.w = doc["bgs"][i]["frames"][j]["w"].GetInt();
					frame.h = doc["bgs"][i]["frames"][j]["h"].GetInt();
					frame.ox = (doc["bgs"][i]["frames"][j].HasMember("ox")) ? doc["bgs"][i]["frames"][j]["ox"].GetInt() : 0;
					frame.oy = (doc["bgs"][i]["frames"][j].HasMember("oy")) ? doc["bgs"][i]["frames"][j]["oy"].GetInt() : 0;
					frame.delay = doc["bgs"][i]["frames"][j]["delay"].GetFloat();
					bg.anim.frames.push_back(frame);
				}
			}
			else
			{
				BgFrame frame;
				frame.x = 0;
				frame.y = 0;
				frame.w = -1; // will be set to image sizeX
				frame.h = -1; // will be set to image sizeY
				frame.ox = 0;
				frame.oy = 0;
				frame.delay = 600;
				bg.anim.frames.push_back(frame);
			}
		}
		else if (doc["bgs"][i].HasMember("rect"))
		{
			bg.isRect = true;
			bg.rect.w = doc["bgs"][i]["rect"]["w"].GetInt();
			bg.rect.h = doc["bgs"][i]["rect"]["h"].GetInt();
			for (int j=0; j<4; j++)
			{
				int colors[3] = {doc["bgs"][i]["rect"]["colors"][j][0].GetInt(), doc["bgs"][i]["rect"]["colors"][j][1].GetInt(), doc["bgs"][i]["rect"]["colors"][j][2].GetInt()};
				bg.rect.colors[j] = Color::new256(colors[0], colors[1], colors[2]);
			}

			BgFrame frame;
			frame.x = 0;
			frame.y = 0;
			frame.w = bg.rect.w;
			frame.h = bg.rect.h;
			frame.ox = 0;
			frame.oy = 0;
			frame.delay = 600;
			bg.anim.frames.push_back(frame);
		}

		bg.bordercolor.on = doc["bgs"][i].HasMember("bordercolor");
		bg.bordercolor.top = bg.bordercolor.bottom = bg.bordercolor.left = bg.bordercolor.right = -1;
		if (bg.bordercolor.on)
		{
			if (doc["bgs"][i]["bordercolor"].IsArray())
			{
				int col = Color::new256(doc["bgs"][i]["bordercolor"][0].GetInt(), doc["bgs"][i]["bordercolor"][1].GetInt(), doc["bgs"][i]["bordercolor"][2].GetInt());
				bg.bordercolor.top = bg.bordercolor.bottom = bg.bordercolor.left = bg.bordercolor.right = col;
			}
			else if (doc["bgs"][i]["bordercolor"].IsObject())
			{
				if (bg.tileX && !bg.tileY)
				{
					if (doc["bgs"][i]["bordercolor"].HasMember("top")) bg.bordercolor.top = Color::new256(doc["bgs"][i]["bordercolor"]["top"][0].GetInt(), doc["bgs"][i]["bordercolor"]["top"][1].GetInt(), doc["bgs"][i]["bordercolor"]["top"][2].GetInt());;
					if (doc["bgs"][i]["bordercolor"].HasMember("bottom")) bg.bordercolor.bottom = Color::new256(doc["bgs"][i]["bordercolor"]["bottom"][0].GetInt(), doc["bgs"][i]["bordercolor"]["bottom"][1].GetInt(), doc["bgs"][i]["bordercolor"]["bottom"][2].GetInt());;
				}
				else if (!bg.tileX && bg.tileY)
				{
					if (doc["bgs"][i]["bordercolor"].HasMember("left")) bg.bordercolor.left = Color::new256(doc["bgs"][i]["bordercolor"]["left"][0].GetInt(), doc["bgs"][i]["bordercolor"]["left"][1].GetInt(), doc["bgs"][i]["bordercolor"]["left"][2].GetInt());;
					if (doc["bgs"][i]["bordercolor"].HasMember("right")) bg.bordercolor.right = Color::new256(doc["bgs"][i]["bordercolor"]["right"][0].GetInt(), doc["bgs"][i]["bordercolor"]["right"][1].GetInt(), doc["bgs"][i]["bordercolor"]["right"][2].GetInt());;
				}
			}
		}

		stage.bgs.push_back(bg);
	}

	delete[] buf;
}

void Engine::setFighter(int ind, std::string name, bool isLocal)
{
	if (started) return;

	char buf[256];
	sprintf(buf, "Set player %d to %s", ind+1, name.c_str());
	printMsg(buf);
	players[ind]->setFighter(name, isLocal);
}

void Engine::playSnd(std::string name)
{
	if (!started) return;

	if (snds.count(name)) wav_play_handle(snds[name], 2);
}

void Engine::setTimeScale(float scale, int autoreset, Callback newAction)
{
	if (!started) return;

	timeScale = scale;
	timeAutoReset = autoreset;
	timeCallback = newAction;
}

void Engine::moveToFront(int ind)
{
	if (!started) return;

	Player* who = players[ind];
	Player* front = playerDrawOrder[1];
	if (who == front) return;

	playerDrawOrder[0] = front; // to back
	playerDrawOrder[1] = who; // to front
}

void Engine::start()
{
	if (started) return;

	char buf[256];

	printMsg("Loading default sounds");
	std::vector<FILELIST> soundfiles = System::listDirectory("/data/smbz-ds/sound", true);
	for (u32 i=0; i<soundfiles.size(); i++)
	{
		u32 len = soundfiles[i].name.size();
		sprintf(buf, "/data/smbz-ds/sound/%s", soundfiles[i].name.c_str());

		snds[soundfiles[i].name.substr(0, len-4)] = wav_load_handle(buf);
	}

	sprintf(buf, "Loading stage \"%s\"", stage.name.c_str());
	printMsg(buf);

	for (u32 i=0; i<stage.bgs.size(); i++)
	{
		BgInfo* bg = &stage.bgs[i];
		
		if (bg->filename.size() > 0)
		{
			sprintf(buf, "/data/smbz-ds/stages/%s/%s", stage.name.c_str(), bg->filename.c_str());
			bg->image = Image::load(buf, TO_RAM);
			if (bg->anim.frames.size() == 1)
			{
				bg->anim.frames[0].w = bg->image->sizeX;
				bg->anim.frames[0].h = bg->image->sizeY;
			}
		}
		else // rect
			bg->image = NULL;
	}

	sprintf(buf, "/data/smbz-ds/stages/%s/ground.png", stage.name.c_str());
	stage.ground.image = (stage.ground.on) ? Image::load(buf, TO_VRAM) : NULL;

	int startX = -32;
	int endX = 256+32;
	int step = (endX-startX) / 2;
	for (int i=0; i<2; i++)
	{
		sprintf(buf, "Loading player %d %s", i+1, players[i]->getInfo()->name.c_str());
		printMsg(buf);

		players[i]->init(150, (i>0 && players[i-1]->getInfo()->name == players[i]->getInfo()->name));
		players[i]->setPos(step*(i+1), -96);
		playerDrawOrder[i] = players[i];

		camFollow.push_back(players[i]);
	}
	for (int i=0; i<2; i++) players[i]->facePlayer(players[1-i]);

	fadeAlpha = 100;
	fadeAdd = -1;
	darkbgAlpha = 0;
	darkbgAdd = 0;
	gameOverTick = 60*5;
	countDownTicks[0] = 60*2; // ready
	countDownTicks[1] = 30; // fight
	timeScale = 1;
	timeAutoReset = 0;
	timeCallback = NULL;

	sprintf(buf, "/data/smbz-ds/stages/%s/intro.mp3", stage.name.c_str());
	if (System::fileExists(buf))
		mp3_play(buf);
	else
	{
		sprintf(buf, "/data/smbz-ds/stages/%s/loop.mp3", stage.name.c_str());
		mp3_play(buf, 1);
	}

	camX = 0;
	camY = 0;
	camZoom = 1;

	ticks = 0;
	started = true;
}

void Engine::stop()
{
	if (!started) return;
	mp3_stop();

	char buf[256];

	std::vector<FILELIST> soundfiles = System::listDirectory("/data/smbz-ds/sound", true);
	for (u32 i=0; i<soundfiles.size(); i++)
	{
		u32 len = soundfiles[i].name.size();
		sprintf(buf, "/data/smbz-ds/sound/%s", soundfiles[i].name.c_str());

		wav_free_handle(snds[soundfiles[i].name.substr(0, len-4)]);
	}
	snds.clear();

	if (stage.ground.on) Image::destroy(stage.ground.image);
	for (u32 i=0; i<stage.bgs.size(); i++)
	{
		if (stage.bgs[i].image) Image::destroy(stage.bgs[i].image);
	}

	for (int i=0; i<2; i++)
	{
		players[i]->destroy();
	}

	started = false;
}

void Engine::tick()
{
	if (!started) return;
	float dt = timeScale;

	char buf[128];
	if (!mp3_is_playing())
	{
		sprintf(buf, "/data/smbz-ds/stages/%s/loop.mp3", stage.name.c_str());
		mp3_play(buf, 1);
	}

	/*if (Key.held.L && camZoom > 0.3f) camZoom -= 0.1f;
	else if (Key.held.R && camZoom < 1) camZoom += 0.1f;
	else if (Key.newPress.Select) camZoom = 1;
	if (Key.held.Up) camY-=(!Key.held.B) ? 16 : 1;
	if (Key.held.Down) camY+=(!Key.held.B) ? 16 : 1;
	if (Key.held.Left) camX-=(!Key.held.B) ? 16 : 1;
	if (Key.held.Right) camX+=(!Key.held.B) ? 16 : 1;*/

	if (Key.newPress.Start) fadeAdd = 1;

	if (fadeAdd)
	{
		fadeAlpha += fadeAdd;
		if (fadeAlpha > 100)
		{
			// round over
			stop();
			return;
		}
		else if (fadeAlpha < 0) fadeAdd = 0;
	}
	if (darkbgAdd)
	{
		darkbgAlpha += darkbgAdd;
		if (darkbgAlpha >= 50 || darkbgAlpha <= 0) darkbgAdd = 0;
	}

	if (timeAutoReset)
	{
		timeAutoReset--;
		if (timeAutoReset <= 0)
		{
			timeScale = 1;
			//if engine.time.callback) engine.time.callback() end
			if (timeCallback) timeCallback((void*)this);
		}
	}

	for (u32 i=0; i<stage.bgs.size(); i++)
	{
		BgInfo* bg = &stage.bgs[i];
		BgAnim* anim = &bg->anim;
		BgFrame* frame = &anim->frames[anim->frame];

		if (!bg->isRect)
		{
			anim->delay += dt;
			while (anim->delay > frame->delay)
			{
				anim->delay -= frame->delay;
				anim->frame++;
				if (anim->frame >= anim->frames.size())
					anim->frame = 0;
				frame = &anim->frames[anim->frame];
			}
		}

		bg->scrollPosX += bg->scrollX * dt;
		bg->scrollPosY += bg->scrollY * dt;
		if (bg->scrollPosX < -bg->image->sizeX) bg->scrollPosX += bg->image->sizeX;
		else if (bg->scrollPosX > bg->image->sizeX) bg->scrollPosX -= bg->image->sizeX;
		if (bg->scrollPosY < -bg->image->sizeY) bg->scrollPosY += bg->image->sizeY;
		else if (bg->scrollPosY > bg->image->sizeY) bg->scrollPosY -= bg->image->sizeY;
	}

	for (int i=0; i<2; i++)
	{
		players[i]->tick(dt);
		if (players[i]->getHP() <= 0) // it's over
		{
			gameOverTick--;
			if (gameOverTick <= 0) fadeAdd = 1;
		}
	}

	float totalpos[2] = {0,0};
	int leftmost = -1;
	int rightmost = -1;
	for (u32 i=0; i<camFollow.size(); i++)
	{
		if (leftmost < 0 || camFollow[i]->getPosX() <= camFollow[leftmost]->getPosX()) leftmost = (int)i;
		if (rightmost < 0 || camFollow[i]->getPosX() >= camFollow[rightmost]->getPosX()) rightmost = (int)i;
		totalpos[0] += camFollow[i]->getPosX();
		totalpos[1] += camFollow[i]->getPosY();
	}
	totalpos[0] /= camFollow.size();
	totalpos[1] /= camFollow.size();
	camX = floor(totalpos[0]);
	camY = floor(totalpos[1])-32;

	float dist = abs(camFollow[rightmost]->getPosX() - camFollow[leftmost]->getPosX());
	camZoom = (dist > 208) ? 208.f/dist : 1;

	if (fightStarted()) ticks += dt;
}

void Engine::draw()
{
	if (!started) return;
	float x, x2, y, screenW, screenH, zoom;
	int w, h, maxColumns, maxRows;

	if (stage.bgcolor > 0) screen::drawFillRect(SCREEN_UP, 0, 0, 256, 192, stage.bgcolor);

	for (u32 i=0; i<stage.bgs.size(); i++)
	{
		BgInfo* bg = &stage.bgs[i];
		BgAnim* anim = &bg->anim;
		BgFrame* frame = &anim->frames[anim->frame];
		w = frame->w;
		h = frame->h;

		zoom = (!bg->scale) ? 1 : camZoom;
		screen::resetView();
		screen::scaleView(SCREEN_UP, zoom, zoom);
		screenW = 256/zoom;
		screenH = 192/zoom;

		maxColumns = screenW / w;
		maxRows = screenH / h;
		if (!maxColumns) maxColumns = 1;
		if (!maxRows) maxRows = 1;

		if (bg->tileX || bg->tileY)
		{
			if (bg->tileX)
				x = (int)(((-camX) * bg->paraX)) % (w + bg->sepX) + (screenW/2) - w - bg->offsetX + bg->scrollPosX;
			else
				x = ((-camX) * bg->paraX) + (screenW/2) - w - bg->offsetX + bg->scrollPosX;
			if (bg->tileY)
				y = (int)(((-camY - 96) * bg->paraY)) % (h + bg->sepY) + (screenH/2) - h - bg->offsetY + bg->scrollPosY;
			else
				y = ((-camY - 96) * bg->paraY) + (screenH/2) - h - bg->offsetY + bg->scrollPosY;
		}
		else
		{
			x = ((-camX - 128) * bg->paraX) - bg->offsetX + bg->scrollPosX;
			y = ((-camY - 96) * bg->paraY) - bg->offsetY + bg->scrollPosY;
		}

		if (bg->bordercolor.on)
		{
			if (bg->tileX && !bg->tileY)
			{
				float yStart, yEnd;
				if (y > 0 && bg->bordercolor.top >= 0)
				{
					yStart = 0;
					yEnd = y;
					screen::drawFillRect(SCREEN_UP, 0, yStart, screenW+1, yEnd, bg->bordercolor.top);
				}
				if (y+h < screenH && bg->bordercolor.bottom >= 0)
				{
					yStart = y+h;
					yEnd = screenH;
					screen::drawFillRect(SCREEN_UP, 0, yStart, screenW+1, yEnd, stage.bgs[i].bordercolor.bottom);
				}
			}
			else if (bg->tileY && !bg->tileX)
			{
				float xStart, xEnd;
				if (x > 0 && bg->bordercolor.left >= 0)
				{
					xStart = 0;
					xEnd = x;
					screen::drawFillRect(SCREEN_UP, xStart, 0, xEnd, screenH+1, bg->bordercolor.left);
				}
				if (x+w < screenW && bg->bordercolor.right >= 0)
				{
					xStart = x+w;
					xEnd = screenW;
					screen::drawFillRect(SCREEN_UP, xStart, 0, xEnd, screenH+1, bg->bordercolor.right);
				}
			}
		}

		if (bg->isRect)
			screen::drawGradientRect(SCREEN_UP, x, y, x+w, y+h, bg->rect.colors[0], bg->rect.colors[1], bg->rect.colors[2], bg->rect.colors[3]);
		else
			screen::blit(SCREEN_UP, x, y, bg->image, frame->x, frame->y, frame->w, frame->h);

		if (bg->tileX && bg->tileY)
		{
			for (float lx=x; lx>-w; lx-=w+bg->sepX)
			{
				for (float uy=y; uy>-h; uy-=h+bg->sepY)
				{
					if (bg->isRect)
						screen::drawGradientRect(SCREEN_UP, lx, uy, lx+w, uy+h, bg->rect.colors[0], bg->rect.colors[1], bg->rect.colors[2], bg->rect.colors[3]);
					else
						screen::blit(SCREEN_UP, lx, uy, bg->image, frame->x, frame->y, frame->w, frame->h);
				}
				for (float dy=y+h+bg->sepY; dy<screenH; dy+=h+bg->sepY)
				{
					if (bg->isRect)
						screen::drawGradientRect(SCREEN_UP, lx, dy, lx+w, dy+h, bg->rect.colors[0], bg->rect.colors[1], bg->rect.colors[2], bg->rect.colors[3]);
					else
						screen::blit(SCREEN_UP, lx, dy, bg->image, frame->x, frame->y, frame->w, frame->h);
				}
			}
			for (float rx=x+w+bg->sepX; rx<screenW; rx+=w+bg->sepX)
			{
				for (float uy=y; uy>-h; uy-=h+bg->sepY)
				{
					if (bg->isRect)
						screen::drawGradientRect(SCREEN_UP, rx, uy, rx+w, uy+h, bg->rect.colors[0], bg->rect.colors[1], bg->rect.colors[2], bg->rect.colors[3]);
					else
						screen::blit(SCREEN_UP, rx, uy, bg->image, frame->x, frame->y, frame->w, frame->h);
				}
				for (float dy=y+h+bg->sepY; dy<screenH; dy+=h+bg->sepY)
				{
					if (bg->isRect)
						screen::drawGradientRect(SCREEN_UP, rx, dy, rx+w, dy+h, bg->rect.colors[0], bg->rect.colors[1], bg->rect.colors[2], bg->rect.colors[3]);
					else
						screen::blit(SCREEN_UP, rx, dy, bg->image, frame->x, frame->y, frame->w, frame->h);
				}
			}
		}

		else if (bg->tileX)
		{
			for (float lx=x; lx>-w; lx-=w+bg->sepX)
			{
				if (bg->isRect)
					screen::drawGradientRect(SCREEN_UP, lx, y, lx+w, y+h, bg->rect.colors[0], bg->rect.colors[1], bg->rect.colors[2], bg->rect.colors[3]);
				else
					screen::blit(SCREEN_UP, lx, y, bg->image, frame->x, frame->y, frame->w, frame->h);
			}
			for (float rx=x; rx<screenW; rx+=w+bg->sepX)
			{
				if (bg->isRect)
					screen::drawGradientRect(SCREEN_UP, rx, y, rx+w, y+h, bg->rect.colors[0], bg->rect.colors[1], bg->rect.colors[2], bg->rect.colors[3]);
				else
					screen::blit(SCREEN_UP, rx, y, bg->image, frame->x, frame->y, frame->w, frame->h);
			}
		}
		else if (bg->tileY)
		{
			for (float uy=y; uy>-h; uy-=h+bg->sepY)
			{
				if (bg->isRect)
					screen::drawGradientRect(SCREEN_UP, x, uy, x+w, uy+h, bg->rect.colors[0], bg->rect.colors[1], bg->rect.colors[2], bg->rect.colors[3]);
				else
					screen::blit(SCREEN_UP, x, uy, bg->image, frame->x, frame->y, frame->w, frame->h);
			}
			for (float dy=y; dy<screenH; dy+=h+bg->sepY)
			{
				if (bg->isRect)
					screen::drawGradientRect(SCREEN_UP, x, dy, x+w, dy+h, bg->rect.colors[0], bg->rect.colors[1], bg->rect.colors[2], bg->rect.colors[3]);
				else
					screen::blit(SCREEN_UP, x, dy, bg->image, frame->x, frame->y, frame->w, frame->h);
			}
		}
	}

	screen::resetView();
	screen::scaleView(SCREEN_UP, camZoom, camZoom);
	screenW = 256/camZoom;
	screenH = 192/camZoom;

	if (stage.ground.on)
	{
		float groundpos = screenH/2 - camY - stage.ground.offsetY;
		if (groundpos < screenH) // ground is in view
		{
			x = -fmod(camX, stage.ground.topFrame.w) + (screenW/2) - stage.ground.topFrame.w;
			x2 = -fmod(camX, stage.ground.loopFrame.w) + (screenW/2) - stage.ground.loopFrame.w;

			screen::blit(SCREEN_UP, x, groundpos, stage.ground.image, stage.ground.topFrame.x, stage.ground.topFrame.y, stage.ground.topFrame.w, stage.ground.topFrame.h);
			for (float lx=x; lx>-stage.ground.topFrame.w; lx-=stage.ground.topFrame.w) screen::blit(SCREEN_UP, lx, groundpos, stage.ground.image, stage.ground.topFrame.x, stage.ground.topFrame.y, stage.ground.topFrame.w, stage.ground.topFrame.h);
			for (float rx=x; rx<screenW; rx+=stage.ground.topFrame.w) screen::blit(SCREEN_UP, rx, groundpos, stage.ground.image, stage.ground.topFrame.x, stage.ground.topFrame.y, stage.ground.topFrame.w, stage.ground.topFrame.h);

			// looping ground
			y = groundpos + stage.ground.topFrame.h;
			while (y < screenH)
			{
				screen::blit(SCREEN_UP, x2, y, stage.ground.image, stage.ground.loopFrame.x, stage.ground.loopFrame.y, stage.ground.loopFrame.w, stage.ground.loopFrame.h);
				for (float lx=x2; lx>-stage.ground.loopFrame.w; lx-=stage.ground.loopFrame.w) screen::blit(SCREEN_UP, lx, y, stage.ground.image, stage.ground.loopFrame.x, stage.ground.loopFrame.y, stage.ground.loopFrame.w, stage.ground.loopFrame.h);
				for (float rx=x2; rx<screenW; rx+=stage.ground.loopFrame.w) screen::blit(SCREEN_UP, rx, y, stage.ground.image, stage.ground.loopFrame.x, stage.ground.loopFrame.y, stage.ground.loopFrame.w, stage.ground.loopFrame.h);
				y += stage.ground.loopFrame.h;
			}
		}
	}

	if (darkbgAlpha > 0)
	{
		screen::setAlpha(darkbgAlpha);
		screen::drawFillRect(SCREEN_UP, 0, 0, screenW+1, screenH+1, 0);
		screen::setAlpha(100);
	}
	for (int i=1; i>=0; i--)
	{
		playerDrawOrder[i]->draw(screenW, screenH);
	}

	// UI
	screen::resetView();
	screenW = 256; screenH = 192;
	if (fadeAlpha > 0)
	{
		screen::setAlpha(fadeAlpha);
		screen::drawFillRect(SCREEN_UP, 0, 0, screenW, screenH, 0);
		screen::setAlpha(100);
	}

	// timer
	float totaltime = ticks / 60.f;
	float secs = fmod(totaltime, 60);
	int mins = (int)(totaltime / 60);
	char buf[64];
	sprintf(buf, "%02d:%06.3f", mins, secs);
	std::string timer = buf;
	screen::print(SCREEN_UP, 128-(timer.size()*3)+1, 6+1, buf, 0);
	screen::print(SCREEN_UP, 128-(timer.size()*3), 6, buf);

	// player healthbars
	for (int i=0; i<2; i++)
	{
		float hpX1, hpX2, hpX1max, hpX2max, x0,y0,x1,y1,x2,y2,x3,y3, x0max, x1max, x2max, x3max, textX;
		float maxhp = 80;
		float health = players[i]->getHP() / players[i]->getMaxHP() * maxhp;

		if (i>0)
		{
			hpX2 = 256-8;
			hpX1 = hpX2 - health;
			hpX2max = hpX2;
			hpX1max = hpX2 - maxhp;
		}
		else
		{
			hpX1 = 8;
			hpX2 = hpX1 + health;
			hpX1max = hpX1;
			hpX2max = hpX1 + maxhp;
		}

		x0 = hpX1;			y0 = 4;
		x1 = hpX1;			y1 = 4+12;
		x2 = hpX2;			y2 = 4+12;
		x3 = hpX2;			y3 = 4;
		x0max = hpX1max-2;
		x1max = hpX1max-2;
		x2max = hpX2max+2;
		x3max = hpX2max+2;

		int c1 = RGB15(0,31,0);
		int c2 = RGB15(0,24,0);
		int playercolor = (i>0) ? RGB15(24,0,0) : RGB15(0,0,24);

		char _text[32];
		sprintf(_text, "%.2f", players[i]->getHP());
		std::string text(_text);
		textX = (i>0) ? hpX2-(text.size()*6)-4 : hpX1+4;

		skew(&x0, &x3, &x1, &x2, (i>0) ? -5 : 5);
		skew(&x0max, &x3max, &x1max, &x2max, (i>0) ? -6 : 6);

		screen::drawQuad(SCREEN_UP, x0max,y0-2,x1max,y1+2,x2max,y2+2,x3max,y3-2, 0,0,0,0); // black outline
		if (players[i]->getHP() > 0 && floor(players[i]->getHP())+1 > 0) screen::drawQuad(SCREEN_UP, x0,y0,x1,y1,x2,y2,x3,y3, c1,c2,c2,c1); // the bar itself
		screen::print(SCREEN_UP, textX+1, (y0+y1)/2-4+1, _text, 0); // health text outline
		screen::print(SCREEN_UP, textX, (y0+y1)/2-4, _text); // health text

		u32 size = players[i]->getInfo()->name.size();
		int nameX = (i>0) ? 256-(size*6)-6 : 6;
		screen::drawFillRect(SCREEN_UP, nameX-4, y1+2, nameX+(size*6)+4, y1+16, playercolor);
		screen::drawFillRect(SCREEN_UP, nameX-2, y1+4, nameX+(size*6)+2, y1+14, 0);
		screen::print(SCREEN_UP, nameX, y1+5, players[i]->getInfo()->name.c_str());
	}

	if (countDownTicks[0] > 0)
	{
		screen::scaleView(SCREEN_UP, 4, 4);
		std::string txt = "Ready...";
		screen::print(SCREEN_UP, 128/4-(txt.size()*(3))+1, 96/4-(4)+1, txt.c_str(), 0);
		screen::print(SCREEN_UP, 128/4-(txt.size()*(3)), 96/4-(4), txt.c_str());
		screen::resetView();
		countDownTicks[0]--;
	}
	else if (countDownTicks[1] > 0)
	{
		screen::scaleView(SCREEN_UP, 4, 4);
		std::string txt = "Fight!";
		screen::print(SCREEN_UP, 128/4-(txt.size()*(3))+1, 96/4-(4)+1, txt.c_str(), 0);
		screen::print(SCREEN_UP, 128/4-(txt.size()*(3)), 96/4-(4), txt.c_str());
		screen::resetView();
		countDownTicks[1]--;
	}
}